import ClaraConstants from './clara-constants';

export default class ClaraCamera {

  lookAtNodeId = null;
  mainCameraId = null;

  claraApi = null;
  scene = null;
  events = null;

  constructor(claraApi, scene, events) {
    this.claraApi = claraApi;
    this.scene = scene;
    this.events = events;
    this.events.registerOnLoaded(this._onClaraSceneLoaded);
  }

  _onClaraSceneLoaded = () => {
    this.lookAtNodeId = this.scene.createEmptyNode('LOOK_AT_NODE');
  };

  getMainCameraId = () => {
    if (!this.mainCameraId) {
      this.mainCameraId = this.claraApi.scene.find({ type: ClaraConstants.CAMERA, name: ClaraConstants.CAMERA });
    }
    return this.mainCameraId;
  };

  setupEasingCamera = () => {
    const mainCameraPosition = this.scene.getPosition(this.getMainCameraId());
    const easingStartPosition = {
      x: mainCameraPosition.z * -2,
      y: mainCameraPosition.y,
      z: mainCameraPosition.x
    };
    const easingTargetId = this.scene.createEmptyNode('EASING_TARGET');
    const easingCameraId = this.createCamera('EASING_CAMERA', easingStartPosition, easingTargetId);
    this.useCamera(easingCameraId);
    return easingCameraId;
  };

  createCamera = (name, position, lookAtTargetId) => {
    const plugs = {
      Transform: [
        [ClaraConstants.TRANSFORM, {
          translation: position
        }],
        ['LookAt', {
          target: lookAtTargetId
        }]
      ],
      Camera: [
        [ClaraConstants.CAMERA, {
          nearClip: 0.01
        }]
      ]
    };
    return this.scene.createOnuNodeByType(name, ClaraConstants.CAMERA, this.scene.getObjectsRootId(), plugs);
  };

  useCamera = cameraId => {
    this.claraApi.player.useCamera(cameraId);
  };

  animateToMainCamera = time => {
    return this.animateToCamera(this.getMainCameraId(), time);
  };

  animateToCamera = (cameraId, time) => {
    return this.claraApi.player.animateCameraTo(cameraId, time);
  };

  flyTo = (nodeId, zoomLevel) => {
    const zoomLevelValue = !zoomLevel ? this.getAutoCalculatedZoomLevel(nodeId) : zoomLevel;
    const nodeCenter = this.scene.getCenter(nodeId);
    const lookAtCameraId = this.createSecondaryLookAtCamera(nodeCenter, zoomLevelValue);
    this.animateToCamera(lookAtCameraId, 2000)
      .then(() => {
        this.claraApi.sceneGraph.deleteNode(lookAtCameraId);
      });
  };

  getAutoCalculatedZoomLevel = nodeId => {
    const boundingLength = this.scene.getAverageBoundingBoxLength(nodeId);
    return 0.0995359 + (2.906155 * boundingLength) + (-0.7061069 * boundingLength * boundingLength);
  };

  createSecondaryLookAtCamera = (partCenter, zoomLevel) => {
    this.scene.setPosition(this.lookAtNodeId, partCenter);
    const cameraPosition = this.calculateCameraPositionToViewNode(this.lookAtNodeId, zoomLevel);
    return this.createCamera('SECONDARY_CAMERA', cameraPosition, this.lookAtNodeId);
  };

  calculateCameraPositionToViewNode = (nodeId, zoomLevel) => {
    const nodeTranslationWorld = this.scene.getTranslationInWorld(this.lookAtNodeId);
    const originToNodeVectorNormalized = nodeTranslationWorld.clone().normalize();
    const cameraVector = originToNodeVectorNormalized.multiplyScalar(zoomLevel);
    return nodeTranslationWorld.add(cameraVector);
  };

}
