import ClaraScene from './clara-scene';
import ClaraCamera from './clara-camera';
import ClaraAnimation from './clara-animation';
import ClaraEvents from './clara-events';
import ColorHelper from './color-helper';
import ClaraConstants from './clara-constants';
import ClaraSnapshot from './clara-snapshot';

export default class ClaraFacade {

  sceneContainerId = null;
  claraApi = null;

  events = null;
  scene = null;
  animation = null;
  snapshot = null;
  camera = null;

  outOfFocusMaterialId = null;
  initialMeshMaterials = null;

  initializeScene = sceneContainerId => {
    this.sceneContainerId = sceneContainerId;
    this.claraApi = claraplayer(this.sceneContainerId); // eslint-disable-line no-undef
    this.claraApi.player.displayThumbnail(false);
    this.claraApi.selection.setHighlighting(true);
    this.claraApi.commands.updateCommand('select', {enabled: false});

    this.events = new ClaraEvents(this.claraApi, this.sceneContainerId);
    this.scene = new ClaraScene(this.claraApi);
    this.camera = new ClaraCamera(this.claraApi, this.scene, this.events);
    this.animation = new ClaraAnimation(this.claraApi);
    this.snapshot = new ClaraSnapshot(this.claraApi);

    this._registerEvents();
  };

  _registerEvents = () => {
    this.events.registerOnPreloaded(this._onClaraScenePreloaded);
    this.events.registerOnHomeButtonPressed(this.animation.resetAllAnimations);
  };

  _onClaraScenePreloaded = () => {
    const baseColor = { r: 1, g: 0.5, b: 0.5 };
    this.outOfFocusMaterialId = this.scene.createMaterial('OUT_OF_FOCUS_MAT', { opacityFactor: 0.5, baseColor });
    const meshMaterialsReference = {
      type: ClaraConstants.POLYMESH,
      plug: ClaraConstants.MATERIAL,
      property: ClaraConstants.REFERENCE
    };
    this.initialMeshMaterials = this.claraApi.scene.getAll(meshMaterialsReference);
  };

  renderScene = claraSceneId => {
    this.claraApi.sceneIO.fetchAndUse(claraSceneId, null, {waitForPublish: true});
  };

  renderPublishedScene = (claraSceneId, hash) => {
    this.claraApi.sceneIO.fetchAndUse(claraSceneId, hash);
  };

  disableTools = tools => {
    tools.forEach(tool => this.disableTool(tool));
  };

  hideTools = tools => {
    tools.forEach(tool => this.hideTool(tool));
  };

  disableTool = tool => {
    this.claraApi.player.removeTool(tool);
  };

  hideTool = tool => {
    this.claraApi.player.hideTool(tool);
  };

  hideToolbar = () => {
    Object.values(ClaraConstants.toolbar).forEach(tool => this.hideTool(tool));
  };

  enableHelpTool = onShowHelp => {
    this.addExecuteToolbarButton('help', () => {
      this.claraApi.player.exitFullscreen();
      onShowHelp();
    });
  };

  setActiveColor = rgbColor => {
    this.claraApi.player.setActiveColor(rgbColor);
  };

  setHoverColor = rgbColor => {
    this.claraApi.player.setHoverColor(rgbColor);
  };

  showMesh = name => {
    const nodeId = this.scene.findMeshOrNullByName(name);

    if (!nodeId) {
      console.error(`Cannot find node named: ${name}`);
      return;
    }

    this.hideAllSiblingMeshes(nodeId);
  };

  hideMesh = name => {
    const nodeId = this.scene.findMeshOrNullByName(name);
    this.setVisibilityRecursively(nodeId, false);
  };

  setVisibility = (nodeId, isVisible = true) => {
    this.claraApi.scene.set({
      id: nodeId,
      plug: 'Properties',
      property: ClaraConstants.VISIBLE,
    }, isVisible);
  };

  hideAllSiblingMeshes = nodeId => {
    const parentId = this.scene.findParent(nodeId);

    const siblings = this.claraApi.scene.filter({
      from: { id: parentId },
      shallow: true
    });

    siblings.forEach(siblingId => {
      this.setVisibilityRecursively(siblingId, false);
    });
    this.setVisibilityRecursively(nodeId, true);
  };

  spawnAnnotation = (childName, name) => {
    const childId = this.scene.findMeshOrNullByName(childName);

    if (!childId) {
      console.error(`Could not find node with name "${childName}" in order to place a hotspot!`);
      return null;
    }
    return this.scene.createAnnotation(name, childId);
  };

  setAnnotationText = (annotationId, text) => {
    this.claraApi.scene.set({id: annotationId, plug: ClaraConstants.ANNOTATION, property: 'useText'}, true);
    this.claraApi.scene.set({id: annotationId, plug: ClaraConstants.ANNOTATION, property: 'text'}, text);
  };

  setVisibilityRecursively = (nodeId, isVisible) => {
    this.claraApi.scene.setAll({
      from: { id: nodeId },
      plug: 'Properties',
      property: ClaraConstants.VISIBLE,
      includeParent: true
    }, isVisible);
  };

  overrideClaraAnnotations = updateFunctionToCall => {
    this.claraApi.annotations.useCustomAnnotationFunction(updateFunctionToCall);
  };

  changeColor = (materialName, r, g, b) => {
    const materialId = this.scene.findMaterial(materialName);
    if (!materialId) {
      console.error(`Cannot find material named: ${materialName}`);
      return;
    }

    const baseColor = this.claraApi.scene.find({
      id: materialId,
      plug: ClaraConstants.MATERIAL,
      property: ClaraConstants.BASE_COLOR
    });
    this.claraApi.scene.set(baseColor, this.createColorObjectForClara(r, g, b));
  };

  changeBaseTexture = (materialName, textureName) => {
    const materialId = this.scene.findMaterial(materialName);
    if (!materialId) {
      console.error(`Cannot find material named: ${materialName}`);
      return;
    }

    const textureId = this.scene.findTexture(textureName);
    if (!textureId) {
      console.error(`Cannot find texture named: ${textureName}`);
      return;
    }

    const materialBaseMap = { id: materialId, plug: ClaraConstants.MATERIAL, property: ClaraConstants.BASE_MAP };
    this.claraApi.scene.set(materialBaseMap, textureId);
  };

  createColorObjectForClara = (r, g, b) => {
    return ColorHelper.convertColorGammaToLinear(ColorHelper.createColorObject0to1(r, g, b));
  };

  convertPositionOffsetUnityToClara = offset => {
    const result = {...offset};
    result.x *= -100;
    result.y *= 100;
    result.z *= 100;
    return result;
  };

  swapMaterialByMeshName = (meshName, materialName) => {
    const meshNodeId = this.claraApi.scene.find({ type: ClaraConstants.POLYMESH, name: meshName });
    if (!meshNodeId) {
      console.error(`Could not find a mesh with name "${meshName}" in order to swap the material.`);
      return;
    }

    const materialId = this.scene.findMaterial(materialName);
    if (!materialId) {
      console.error(`Could not find material with name "${materialName}' in order to swap the material.`);
      return;
    }
    this.scene.setMaterial(meshNodeId, materialId);
  };

  rotateAroundWorldAxis = (nodeId, rotation) => {
    const THREE = this.claraApi.deps.THREE;
    const oldParentId = this.scene.findParent(nodeId);
    const nodeTransform = this.claraApi.scene.getWorldTransform(nodeId);

    const eulerRotation = new THREE.Euler(rotation.x, rotation.y, rotation.z, 'ZYX');
    const newTransformMatrix = new THREE.Matrix4().makeRotationFromEuler(eulerRotation).multiply(nodeTransform);
    const parentTransform = this.claraApi.scene.getWorldTransform(oldParentId);

    let newRelativeTransform;
    if (parentTransform) {
      const parentWorldInverse = new THREE.Matrix4().getInverse(this.claraApi.scene.getWorldTransform(oldParentId));
      newRelativeTransform = new THREE.Matrix4().multiplyMatrices(parentWorldInverse, newTransformMatrix);
    } else {
      newRelativeTransform = newTransformMatrix;
    }

    const newNodeTranslate = new THREE.Vector3();
    const newNodeRotate = new THREE.Quaternion();
    const newNodeScale = new THREE.Vector3();
    newRelativeTransform.decompose(newNodeTranslate, newNodeRotate, newNodeScale);
    const rotateVec = new THREE.Euler().setFromQuaternion(newNodeRotate, 'ZYX').toVector3().multiplyScalar(180.0 / Math.PI);

    this.scene.setTransform(nodeId, newNodeTranslate, rotateVec, newNodeScale);
  };

  swapMaterials = (oldMaterialName, newMaterialName) => {
    const newMaterialId = this.scene.findMaterial(newMaterialName);
    if (!newMaterialId) {
      console.error(`Could not find material with name "${newMaterialName}" to swap`);
      return;
    }

    const materialPaths = this.scene.findMeshesWithMaterialByName(oldMaterialName);
    if (materialPaths.length === 0) {
      console.error(`Found no meshes that use material "${oldMaterialName}" to swap`);
      return;
    }

    this.claraApi.scene.setAll(materialPaths, newMaterialId);
  };

  addExecuteToolbarButton = (iconName, onClickedCallback) => {
    return this.claraApi.commands.addCommand({
      options: { icon: iconName },
      display: { playerTools: true },
      label: iconName,
      execute: onClickedCallback
    }, iconName);
  };

  addToggleToolbarButton = (iconName, onActivateCallback, onDeactivateCallback) => {
    return this.claraApi.commands.addCommand({
      options: { icon: iconName },
      display: { playerTools: true },
      activate: onActivateCallback,
      deactivate: onDeactivateCallback
    }, iconName);
  };

  focusOn = (nodeName, zoomLevel) => {
    this.unfocusAll();
    const nodeId = this.scene.findMeshOrNullByName(nodeName);
    if (!nodeId) {
      console.error(`Could not find node with name "${nodeName}" to focus on.`);
      return;
    }
    this.focusOnById(nodeId, zoomLevel);
  };

  focusOnById = (nodeId, zoomLevel) => {
    this.camera.flyTo(nodeId, zoomLevel);
    this.claraApi.selection.selectNode(nodeId);
    const meshMaterialReferencePaths = this.claraApi.scene.filter({
      type: ClaraConstants.POLYMESH,
      plug: ClaraConstants.MATERIAL,
      property: ClaraConstants.REFERENCE
    });
    const focusedNodes = this.claraApi.scene.filter({
      from: {id: nodeId},
      includeParent: true
    });
    const contextNodes = this.claraApi.scene.filter({
      from: {name: 'Context'},
      type: ClaraConstants.POLYMESH,
      includeParent: true
    });
    const nonSelectedMeshes = meshMaterialReferencePaths.filter(path => {
      return !focusedNodes.includes(path[0]) && !contextNodes.includes(path[0]);
    });
    // Apply the newly created material to all unfocused meshes.
    this.claraApi.scene.setAll(nonSelectedMeshes, this.outOfFocusMaterialId);
  };

  unfocusAll = () => {
    Object.keys(this.initialMeshMaterials).forEach(meshId => {
      this.scene.setMaterial(meshId, this.initialMeshMaterials[meshId]);
    });
    this.claraApi.selection.deselectAll();
  };

  setMeshQuality = (level = 0) => {
    const meshClientLevels = this.claraApi.scene.filter({type: 'PolyMesh', plug: 'PolyMesh', property: 'clientLevel'});
    const reductionLevel = this.lodToReductionLevel(level);
    this.claraApi.scene.setAll(meshClientLevels, reductionLevel);
  };

  getMeshQuality = () => {
    const currentLevel = this.claraApi.scene.get({plug: 'PolyMesh', property: 'clientLevel'});
    return this.reductionLevelToLod(currentLevel);
  };

  getNumberOfLods = () => {
    return this.claraApi.scene.get({type: 'PolyMesh', plug: 'PolyMesh', property: 'nbLevel'});
  };

  lodToReductionLevel = lod => {
    const numberOfLods = this.getNumberOfLods();
    const levelOffset = 0.01; // so we don't end up on the level border
    return (lod / numberOfLods) + levelOffset;
  };

  reductionLevelToLod = reductionLevel => {
    if (reductionLevel) {
      const numberOfLods = this.getNumberOfLods();
      return reductionLevel >= 1 ? numberOfLods - 1 : Math.floor(numberOfLods * reductionLevel);
    }
    return null;
  };
}
