import ClaraConstants from './clara-constants';

export default class ClaraScene {

  objectsRootId = null;
  materialsRootId = null;
  rootNodeId = null;

  constructor(claraApi) {
    this.claraApi = claraApi;
  }

  getObjectsRootId = () => {
    if (!this.objectsRootId) {
      this.objectsRootId = this.claraApi.scene.find({type: 'Objects'});
    }
    return this.objectsRootId;
  };

  getMaterialsRootId = () => {
    if (!this.materialsRootId) {
      this.materialsRootId = this.claraApi.scene.find({type: 'MaterialLibrary'});
    }
    return this.materialsRootId;
  };

  getRootNodeId = () => {
    if (!this.rootNodeId) {
      this.rootNodeId = this.findRootModelOrNull();
    }
    return this.rootNodeId;
  };

  findParent = nodeId => {
    return this.claraApi.scene.find({
      id: nodeId,
      parent: true
    });
  };

  findRootModelOrNull = () => {
    let rootNodeId = this.findFirstLevelNodeInScene(ClaraConstants.MODEL);

    if (!rootNodeId) {
      rootNodeId = this.findFirstLevelNodeInScene(ClaraConstants.NULL);
    }

    return rootNodeId;
  };

  findFirstLevelNodeInScene = type => {
    return this.claraApi.scene.find({ from: { id: this.getObjectsRootId() }, shallow: true, type });
  };

  findMeshOrNullByName = name => {
    return this.findByNameAndTypes([ClaraConstants.POLYMESH, ClaraConstants.NULL, ClaraConstants.MODEL], name);
  };

  findByNameAndTypes = (types, name) => {
    return this.claraApi.scene.find({type: types, name});
  };

  findMeshesWithMaterialByName = materialName => {
    const materialPaths = [];
    const meshIds = this.claraApi.scene.filter({type: ClaraConstants.POLYMESH});
    meshIds.forEach(meshId => {
      const materialRef = this.getMaterialReference(meshId);
      if (materialRef !== null) {
        const existingMaterial = this.claraApi.scene.get({id: materialRef, property: 'name'});
        if (existingMaterial === materialName) {
          materialPaths.push(this.findMaterialReference(meshId));
        }
      }
    });
    return materialPaths;
  };

  findMaterialReference = meshId => {
    return this.claraApi.scene.find({id: meshId, plug: ClaraConstants.MATERIAL, property: ClaraConstants.REFERENCE});
  };

  findMaterial = name => {
    return this.findByNameAndTypes(ClaraConstants.MATERIAL, name);
  };

  findTexture = name => {
    return this.findByNameAndTypes(ClaraConstants.IMAGE, name);
  };

  getId = node => {
    return node._result; // eslint-disable-line no-underscore-dangle
  };

  getCameraTransform = () => {
    const cameraId = this.claraApi.player.getCamera();
    return this.getTransform(cameraId);
  };

  setTransform = (nodeId, position, rotation, scale) => {
    this.setPosition(nodeId, position);
    this.setRotation(nodeId, rotation);
    this.setScale(nodeId, scale);
  };

  setPosition = (nodeId, position) => {
    this.setTransformProperty(nodeId, position, ClaraConstants.TRANSLATION);
  };

  setRotation = (nodeId, rotation) => {
    this.setTransformProperty(nodeId, rotation, ClaraConstants.ROTATION);
  };

  setScale = (nodeId, scale) => {
    this.setTransformProperty(nodeId, scale, ClaraConstants.SCALE);
  };

  setTransformProperty = (nodeId, value, propertyName) => {
    this.claraApi.scene.set({
      id: nodeId,
      plug: ClaraConstants.TRANSFORM,
      property: propertyName
    }, value);
  };

  getTransform = nodeId => {
    const position = this.getPosition(nodeId);
    const rotation = this.getRotation(nodeId);
    return { position, rotation };
  };

  getPosition = nodeId => this.getTransformProperty(nodeId, ClaraConstants.TRANSLATION);

  getRotation = nodeId => this.getTransformProperty(nodeId, ClaraConstants.ROTATION);

  getScale = nodeId => this.getTransformProperty(nodeId, ClaraConstants.SCALE);

  getTransformProperty = (nodeId, propertyName) => this.claraApi.scene.get({
    id: nodeId,
    plug: ClaraConstants.TRANSFORM,
    property: propertyName
  });

  getTranslationInWorld = nodeId => {
    const THREE = this.claraApi.deps.THREE;
    const partTransformMatrix = this.claraApi.scene.getWorldTransform(nodeId);
    const translation = new THREE.Vector3();
    translation.setFromMatrixPosition(partTransformMatrix);
    return translation;
  };

  getAverageBoundingBoxLength = nodeId => {
    const boundingBox = this.claraApi.scene.getNodeBoundingBox(nodeId);
    const lengthX = boundingBox.max.x - boundingBox.min.x;
    const lengthY = boundingBox.max.y - boundingBox.min.y;
    const lengthZ = boundingBox.max.z - boundingBox.min.z;
    return (lengthX + lengthY + lengthZ) / 3;
  };

  getCenter = nodeId => {
    const THREE = this.claraApi.deps.THREE;
    const boundingBox = this.claraApi.scene.getNodeBoundingBox(nodeId);
    const center = new THREE.Vector3();
    center.x = (boundingBox.min.x + boundingBox.max.x) / 2;
    center.y = (boundingBox.min.y + boundingBox.max.y) / 2;
    center.z = (boundingBox.min.z + boundingBox.max.z) / 2;
    return center;
  };

  getMaterialReference = meshId => {
    return this.claraApi.scene.get({id: meshId, plug: ClaraConstants.MATERIAL, property: ClaraConstants.REFERENCE});
  };

  setMaterial = (meshId, materialId) => {
    const meshReference = {id: meshId, plug: ClaraConstants.MATERIAL, property: ClaraConstants.REFERENCE};
    this.claraApi.scene.set(meshReference, materialId);
  };

  createAnnotation = (name, parentId) => {
    const plugs = {
      Annotation: [[ClaraConstants.ANNOTATION, {
        limitVisilibility: true,
        directionMode: 'awayFromOrigin',
        directionTarget: this.getRootNodeId()
      }]],
    };
    return this.createNodeByType(name, ClaraConstants.ANNOTATION, parentId, plugs);
  };

  createMaterial = (name, properties) => {
    const plugs = {
      Material: [[
        ClaraConstants.PHYSICAL,
        properties
      ]]
    };
    return this.createOnuNodeByType(name, ClaraConstants.MATERIAL, this.getMaterialsRootId(), plugs);
  };

  createEmptyNode = name => {
    return this.createOnuNodeByType(name, ClaraConstants.NULL, this.getObjectsRootId());
  };

  createOnuNodeByType = (name, type, parentId, plugs) => {
    return this.createNodeByType(this.appendOnuPrefix(name), type, parentId, plugs);
  };

  createNodeByType = (name, type, parentId, plugs) => {
    const node = this.claraApi.scene.addNode({
      name,
      type,
      parent: parentId,
      plugs
    });
    return this.getId(node);
  };

  appendOnuPrefix = name => {
    return `ONU_${name}`;
  };
}
